@@ -13,7 +13,7 @@ end
 
 function package:_init ()
   base._init(self)
-  self.class:loadPackage("color")
+  self:loadPackage("color")
 end
 
 function package:registerCommands ()
